@@ -1,6 +1,6 @@
 ;;;; build-gtk.jl -- translate guile-gtk .defs file to rep C code
 ;;;  Copyright (C) 1999 John Harper <john@dcs.warwick.ac.uk>
-;;;  $Id: build-gtk.jl,v 1.8 1999/11/30 22:33:19 john Exp $
+;;;  $Id: build-gtk.jl,v 1.14 2000/04/09 17:32:40 john Exp $
 ;;;
 ;;; This program is free software; you can redistribute it and/or modify
 ;;; it under the terms of the GNU General Public License as published by
@@ -19,14 +19,15 @@
 (provide 'build-gtk)
 
 ;; Notes:
-;;
+
 ;; This assumes that the `sed-fix-defs' sed script has been run over all
 ;; input files (to convert schemey things to their lispy equivalents)
-;;
+
 ;; Todo:
-;;  * make the `import' directive work as in guile-gtk?
 ;;  * doesn't check for `listable' type-property
-;;
+;;  * guile-gtk `struct' and `ptype' types
+;;  * not possible to wrap functions returning vector types
+
 ;; WARNING: This makes some pretty gruesome assumptions. [where?]
 
 
@@ -35,27 +36,32 @@
 ;; Alist of (TYPE ["C-TYPE" | DECL-FUNC] ["REP2GTK" | FROM-REP-FUNC]
 ;;           ["GTK2REP" | TO-REP-FUNC] ["PRED-NAME" | PRED-FUNC]
 ;;	     . OPTION-ALIST)
-;;
+
 ;; The required functions are called as:
-;;
+
 ;;   (DECL-FUNC TYPE TYPE-INFO)
 ;;   (FROM-REP-FUNC OUTPUT-STREAM TYPE "REP-VAR" TYPE-INFO OPTIONS)
 ;;   (TO-REP-FUNC OUTPUT-STREAM TYPE "GTK-VAR" TYPE-INFO OPTIONS)
 ;;   (PRED-FUNC OUTPUT-STREAM TYPE "REP-VAR" TYPE-INFO OPTIONS)
-;;
+
 ;; The options in the OPTION-ALIST may be:
-;;
+
 ;;   (c2args . EMIT-ARG-FUNC)
 ;;   (finish . FINISH-ARG-FUNC)
 ;;   (listable . BOOLEAN)
-;;
+
 ;; with:
-;;
+
 ;;   (EMIT-ARG-FUNC OUTPUT TYPE "GTK-VAR" OPTIONS)
 ;;   (FINISH-ARG-FUNC OUTPUT TYPE "GTK-VAR" "REP-VAR" OPTIONS)
-;;
+
 (defvar gtk-type-alist nil)
 
+(defun define-type (type c-type rep-to-gtk gtk-to-rep type-pred &rest options)
+  (setq gtk-type-alist (cons (list* type c-type rep-to-gtk
+				    gtk-to-rep type-pred options)
+			     gtk-type-alist)))
+
 
 ;; Work variables
 
@@ -175,7 +181,7 @@
 	  (when def
 	    (or (consp def) (error "Definition isn't a list"))
 	    (cond
-	     ((eq (car def) 'import)
+	     ((memq (car def) '(include import))
 	      (let
 		  ((file (open-file (expand-file-name (nth 1 def)
 						      (file-name-directory
@@ -183,8 +189,10 @@
 				    'read)))
 		(or file (error "Can't open input file: %s" (nth 1 def)))
 		(unwind-protect
-		    (let
-			((gtk-importing t))
+		    (if (eq (car def) 'import)
+			(let
+			    ((gtk-importing t))
+			  (parse-gtk file))
 		      (parse-gtk file))
 		  (close-file file))))
 	     ((eq (car def) 'define-enum)
@@ -244,6 +252,8 @@
 		  (if cell
 		      (rplacd cell (nthcdr 2 def))
 		    (setq gtk-functions (cons (cdr def) gtk-functions))))))
+	     ((eq (car def) 'define-type)
+	      (eval def))
 	     ((eq (car def) 'options)
 	      (unless gtk-importing
 		(mapc (lambda (cell)
@@ -280,7 +290,29 @@
   (@ "#include <rep.h>\n")
   (@ "#include \"rep-gtk.h\"\n\n"))
 
-(defun output-footer (output))
+(defun output-footer (output)
+  (let*
+      ((feature (gtk-get-option 'provide gtk-options))
+       (c-feature (and feature (gtk-unhyphenate-name (symbol-name feature))))
+       (init (gtk-get-option 'init-func gtk-options)))
+    (when feature
+      (@ "\n#if rep_INTERFACE >= 9\n")
+      (@ "DEFSYM \(%s, \"%s\"\);\n" c-feature feature)
+      (@ "#endif\n")
+      (@ "\nrepv\nrep_dl_init \(void\)\n{\n")
+      (@ "#if rep_INTERFACE >= 9\n")
+      (@ "  repv s = rep_push_structure \(\"%s\"\);\n" c-feature)
+      (@ "#else\n")
+      (@ "  rep_INTERN \(%s\);\n" c-feature)
+      (@ "#endif\n")
+      (when init
+	(@ "\n  %s \(\);\n\n" init))
+      (@ "#if rep_INTERFACE >= 9\n")
+      (@ "  return rep_pop_structure \(s\);\n")
+      (@ "#else\n")
+      (@ "  return Q%s;\n" c-feature)
+      (@ "#endif\n")
+      (@ "}\n"))))
 
 (defun output-imported-enums (output)
   (when gtk-imported-enums
@@ -394,15 +426,16 @@
 	    (@ "};\n\n"))) gtk-objects))
 
 (defun output-type-info (output)
-  (@ "\f\n/* Vector of all type information */\n\n")
-  (@ "static sgtk_type_info *_type_infos[] = {\n")
-  (mapc (lambda (lst)
-	  (mapc (lambda (type)
-		  (@ "  (sgtk_type_info*)&sgtk_%s_info,\n"
-		     (gtk-canonical-name (symbol-name (car type)))))
-		lst))
-	(list gtk-enums gtk-flags gtk-boxed gtk-objects))
-  (@ "  NULL\n};\n\n"))
+  (when (or gtk-enums gtk-flags gtk-boxed gtk-objects)
+    (@ "\f\n/* Vector of all type information */\n\n")
+    (@ "static sgtk_type_info *_type_infos[] = {\n")
+    (mapc (lambda (lst)
+	    (mapc (lambda (type)
+		    (@ "  (sgtk_type_info*)&sgtk_%s_info,\n"
+		       (gtk-canonical-name (symbol-name (car type)))))
+		  lst))
+	  (list gtk-enums gtk-flags gtk-boxed gtk-objects))
+    (@ "  NULL\n};\n\n")))
 
 (defun output-functions (output)
   (@ "\f\n/* Defuns */\n\n")
@@ -420,16 +453,21 @@
   (@ "\f\n/* Initialisation */\n\n")
   (let
       ((init-func (gtk-get-option 'init-func gtk-options))
-       (other-inits (gtk-get-options 'other-inits gtk-options)))
-    (@ "void\n%s (void)\n{\n" init-func)
-    (@ "  static int done;\n  if (!done)\n    {\n")
-    (@ "      done = 1;\n")
-    (mapc (lambda (func)
-	    (@ "      %s ();\n" func)) other-inits)
-    (@ "      sgtk_register_type_infos (_type_infos);\n")
-    (mapc (lambda (cname)
-	    (@ "      rep_ADD_SUBR(S%s);\n" cname)) (nreverse gtk-subrs))
-    (@ "    \}\n\}\n")))
+       (other-inits (gtk-get-options 'other-inits gtk-options))
+       (extra-init (gtk-get-options 'extra-init-code gtk-options)))
+    (when init-func
+      (@ "void\n%s (void)\n{\n" init-func)
+      (@ "  static int done;\n  if (!done)\n    {\n")
+      (@ "      done = 1;\n")
+      (mapc (lambda (func)
+	      (@ "      %s ();\n" func)) other-inits)
+      (when (or gtk-enums gtk-flags gtk-boxed gtk-objects)
+	(@ "      sgtk_register_type_infos (_type_infos);\n"))
+      (mapc (lambda (cname)
+	      (@ "      rep_ADD_SUBR(S%s);\n" cname)) (nreverse gtk-subrs))
+      (mapc (lambda (code)
+	      (@ "      %s\n" code)) extra-init)
+      (@ "    \}\n\}\n"))))
 
 (defun output-gtk (output)
   (output-header output)
@@ -461,10 +499,21 @@
     (setq type (car type)))
   (nth 1 type))
 
-(defun gtk-inner-type-options (type)
+(defun gtk-composite-type-mode (type)
   (while (consp (car type))
     (setq type (car type)))
-  (nthcdr 2 type))
+  (case (car type)
+    ((ret) 'out)
+    ((fvec) (or (nth 3 type) 'in))
+    (t (or (nth 2 type) 'in))))
+
+(defun gtk-composite-type-len (type)
+  (while (consp (car type))
+    (setq type (car type)))
+  (case (car type)
+    ((ret) 1)
+    ((fvec) (nth 2 type))
+    (t nil)))
 
 (defun gtk-type-info (type)
   (let*
@@ -597,55 +646,66 @@
 (defun output-rep-to-full-callback (output type rep-var typage options)
   (let
       ((protect (gtk-get-option 'protection options)))
-    (if (and (not (eq protect t))
-	     (not (eq protect nil)))
-	(@ "sgtk_protect \(p_%s, %s\)" protect rep-var)
-      (@ "sgtk_protect \(Qt, %s\)" rep-var))))
+    (cond ((eq protect '*result*)
+	   (@ "sgtk_new_protect \(%s\)" rep-var))
+	  ((and (not (eq protect t))
+		(not (eq protect nil)))
+	   (@ "sgtk_protect \(p_%s, %s\)" protect rep-var))
+	  (t
+	   (@ "sgtk_protect \(Qt, %s\)" rep-var)))))
 
 (defun output-full-callback-args (output type var options)
   (@ "0, sgtk_callback_marshal, (gpointer)%s, sgtk_callback_destroy" var))
 
+(defun output-full-callback-finish (output type g-var r-var options)
+  (let
+      ((protect (gtk-get-option 'protection options)))
+    (when (eq protect '*result*)
+      (@ "  sgtk_set_protect \(pr_ret, %s\);\n" g-var))))
+
 (defun output-rep-to-cvec (output type rep-var typage)
   (let*
       ((outer-type (gtk-outer-type type))
        (inner-type (gtk-inner-type type))
        (inner-typage (gtk-type-info inner-type))
-       (decl (gtk-type-decl inner-type inner-typage)))
+       (decl (gtk-type-decl inner-type inner-typage))
+       (mode (gtk-composite-type-mode type)))
     (output-helper inner-type standard-output)
     (@ "sgtk_rep_to_cvec \(%s, %s, sizeof \(%s\)\)"
        rep-var
-       (if (eq outer-type 'ret)
+       (if (eq mode 'out)
 	   "0"
-	 ;; all but `ret' args are `in'
 	 (format nil "_sgtk_helper_fromrep_%s" inner-type))
        decl)))
 
-(defun output-cvec-to-rep ()
-  (error "output-rep-to-cvec is unimplemented"))
+(defun output-cvec-to-rep (output type gtk-var typage)
+  (let*
+      ((outer-type (gtk-outer-type type))
+       (inner-type (gtk-inner-type type))
+       (inner-typage (gtk-type-info inner-type))
+       (decl (gtk-type-decl inner-type inner-typage)))
+    (output-helper inner-type standard-output)
+    (@ "sgtk_cvec_to_rep \(&%s, _sgtk_helper_torep_copy_%s, sizeof \(%s\)\)"
+       gtk-var inner-type decl)))
 
 (defun output-cvec-pred (output type rep-var typage)
   (let*
       ((outer-type (gtk-outer-type type))
        (inner-type (gtk-inner-type type))
-       (inner-typage (gtk-type-info inner-type)))
+       (inner-typage (gtk-type-info inner-type))
+       (mode (gtk-composite-type-mode type))
+       (len (gtk-composite-type-len type)))
     (output-helper inner-type standard-output)
-    (cond ((memq outer-type '(cvec cvecr list slist))
-	   ;; XXX assumes `in' or `inout' types
-	   (@ "sgtk_valid_composite \(%s, _sgtk_helper_valid_%s\)"
-	      rep-var inner-type))
-	  ((memq outer-type '(fvec ret))
-	   (let
-	       ((len (if (eq outer-type 'ret)
-			 1
-		       (car (gtk-inner-type-options type)))))
-	     (@ "sgtk_valid_complen \(%s, %s, %s\)"
-		rep-var (if (eq outer-type 'ret)
-			    ;; ret is `out', so don't check inner validity
-			    "NULL"
-			  (concat "_sgtk_helper_valid_"
-				  (symbol-name inner-type))) len)))
-	  (t
-	   (gtk-warning "Don't know type predicate of %s" type)))))
+    (if len
+	(@ "sgtk_valid_complen \(%s, %s, %s\)"
+	   rep-var
+	   (if (eq mode 'out)
+	       ;; `out', so don't check inner validity
+	       "NULL"
+	     (concat "_sgtk_helper_valid_" (symbol-name inner-type)))
+	   len)
+      (@ "sgtk_valid_composite \(%s, _sgtk_helper_valid_%s\)"
+	 rep-var inner-type))))
 
 (defun output-cvec-args (output type var options)
   (let*
@@ -657,9 +717,7 @@
 	   (@ "%s.count, \(%s*\) %s.vec" var decl var))
 	  ((eq outer-type 'cvecr)
 	   (@ "\(%s*\) %s.vec, %s.count" decl var var))
-	  ((eq outer-type 'fvec)
-	   (@ "\(%s*\) %s.vec" decl var))
-	  ((eq outer-type 'ret)
+	  ((memq outer-type '(fvec ret tvec))
 	   (@ "\(%s*\) %s.vec" decl var))
 	  (t
 	   (gtk-warning "Don't know how to pass type %s" type)))))
@@ -669,13 +727,13 @@
       ((outer-type (gtk-outer-type type))
        (inner-type (gtk-inner-type type))
        (inner-typage (gtk-type-info inner-type))
-       (decl (gtk-type-decl inner-type inner-typage)))
+       (decl (gtk-type-decl inner-type inner-typage))
+       (mode (gtk-composite-type-mode type)))
     (@ "  sgtk_cvec_finish \(&%s, %s, %s, sizeof \(%s\)\);\n"
        gtk-var rep-var
-       (if (eq outer-type 'ret)
-	   ;; only `ret' args are `out'
-	   (format nil "_sgtk_helper_torep_nocopy_%s" inner-type)
-	 "0")
+       (if (eq mode 'in)
+	   "0"
+	 (format nil "_sgtk_helper_torep_nocopy_%s" inner-type))
        decl)))
 
 (defun output-rep-to-list (output type rep-var typage)
@@ -696,8 +754,14 @@
 
 (defun output-list-finish (output type gtk-var rep-var options)
   (let
-      ((outer-type (gtk-outer-type type)))
-    (@ "  sgtk_%s_finish \(%s, %s, 0\);\n" outer-type gtk-var rep-var)))
+      ((outer-type (gtk-outer-type type))
+       (inner-type (gtk-inner-type type))
+       (mode (gtk-composite-type-mode type)))
+    (@ "  sgtk_%s_finish \(%s, %s, %s\);\n"
+       outer-type gtk-var rep-var
+       (if (eq mode 'in)
+	   "0"
+	 (format nil "_sgtk_helper_torep_nocopy_%s" inner-type)))))
 
 
 ;; Function generation
@@ -741,7 +805,8 @@
 
     ;; output any gc roots required
     (mapc (lambda (arg)
-	    (when (gtk-type-prop (gtk-arg-type arg) 'finish)
+	    (when (or (gtk-get-arg-options 'protect-during arg)
+		      (gtk-type-prop (gtk-arg-type arg) 'finish))
 	      (@ "  rep_GC_root gc_%s;\n" (gtk-arg-name arg)))) args)
 
     ;; output arg/ret decls
@@ -823,7 +888,8 @@
 
     ;; initialise gc roots
     (mapc (lambda (arg)
-	    (when (gtk-type-prop (gtk-arg-type arg) 'finish)
+	    (when (or (gtk-get-arg-options 'protect-during arg)
+		      (gtk-type-prop (gtk-arg-type arg) 'finish))
 	      (@ "  rep_PUSHGC \(gc_%s, p_%s\);\n"
 		 (gtk-arg-name arg) (gtk-arg-name arg)))) args)
 
@@ -877,19 +943,7 @@
 	  (setq tem (cdr tem))))
       (@ "\);\n\n"))
 
-    ;; output `finish' options
-    (mapc (lambda (arg)
-	    (let
-		((opt (gtk-type-prop (gtk-arg-type arg) 'finish)))
-	      (when opt
-		(if (functionp opt)
-		    (funcall opt output (gtk-arg-type arg)
-			     (concat "c_" (gtk-arg-name arg))
-			     (concat "p_" (gtk-arg-name arg))
-			     options)
-		  (gtk-warning "finish function %s undefined" opt))))) args)
-
-    ;; output ret conversion & return
+    ;; output ret conversion
     (unless (eq ret 'none)
       (let*
 	  ((typage (gtk-type-info ret))
@@ -904,9 +958,22 @@
 		"Don't know how to convert %s to repv" ret)))
 	(@ ";\n")))
 
+    ;; output `finish' options
+    (mapc (lambda (arg)
+	    (let
+		((opt (gtk-type-prop (gtk-arg-type arg) 'finish)))
+	      (when opt
+		(if (functionp opt)
+		    (funcall opt output (gtk-arg-type arg)
+			     (concat "c_" (gtk-arg-name arg))
+			     (concat "p_" (gtk-arg-name arg))
+			     options)
+		  (gtk-warning "finish function %s undefined" opt))))) args)
+
     ;; pop gc roots
     (mapc (lambda (arg)
-	    (when (gtk-type-prop (gtk-arg-type arg) 'finish)
+	    (when (or (gtk-get-arg-options 'protect-during arg)
+		      (gtk-type-prop (gtk-arg-type arg) 'finish))
 	      (@ "  rep_POPGC;\n"
 		 (gtk-arg-name arg) (gtk-arg-name arg)))) args)
 
@@ -929,12 +996,13 @@
 	      (mapc #'(lambda (field)
 			(output-field-accessors
 			 (car def) field output
-			 (car (cdr (assq 'setter (nthcdr 2 field))))))
+			 (car (cdr (assq 'setter (nthcdr 2 field))))
+			 (car (cdr (assq 'getter (nthcdr 2 field))))))
 		    fields))
 	    (output-type-predicate (car def) output)))
 	type-list))
 
-(defun output-field-accessors (datatype field output &optional settable)
+(defun output-field-accessors (datatype field output &optional settable getter)
   (let*
       ((type (car field))
        (cdatatype (gtk-canonical-name (symbol-name datatype)))
@@ -943,7 +1011,9 @@
 			   type (list (list datatype 'obj)))
 		     output
 		     (lambda (output)
-		       (@ "  cr_ret = c_obj->%s;\n" cfieldname)))
+		       (if getter
+			   (@ "  cr_ret = %s (c_obj);\n" getter)
+			 (@ "  cr_ret = c_obj->%s;\n" cfieldname))))
     (when settable
       (output-function (list (intern (format nil "%s_%s_set"
 					     cdatatype cfieldname))
@@ -1065,57 +1135,90 @@
 
 ;; initialisation
 
-(setq gtk-type-alist
-  (list (list 'type "GtkType" "sgtk_rep_to_type" "sgtk_type_to_rep"
-	      "sgtk_valid_type")
-	(list 'char "gchar" "sgtk_rep_to_char" "sgtk_char_to_rep"
-	      "sgtk_valid_char")
-	(list 'bool "int" "sgtk_rep_to_bool" "sgtk_bool_to_rep" nil)
-	(list 'int "gint" "sgtk_rep_to_int" "sgtk_int_to_rep"
-	      "sgtk_valid_int" '(listable . t))
-	(list 'uint "guint" "sgtk_rep_to_uint" "sgtk_uint_to_rep"
-	      "sgtk_valid_uint" '(listable . t))
-	(list 'long "glong" "sgtk_rep_to_long" "sgtk_long_to_rep"
-	      "sgtk_valid_long")
-	(list 'ulong "gulong" "sgtk_rep_to_ulong" "sgtk_ulong_to_rep"
-	      "sgtk_valid_ulong")
-	(list 'float "gfloat" "sgtk_rep_to_float" "sgtk_float_to_rep"
-	      "sgtk_valid_float")
-	(list 'string "char*" "sgtk_rep_to_string" "sgtk_string_to_rep"
-	      "sgtk_valid_string" '(listable . t))
-	(list 'enum output-complex-type output-rep-to-enum output-enum-to-rep
-	      output-enum-pred)
-	(list 'flags output-complex-type output-rep-to-flags
+(define-type 'type "GtkType" "sgtk_rep_to_type"
+	     "sgtk_type_to_rep" "sgtk_valid_type")
+
+(define-type 'char "gchar" "sgtk_rep_to_char"
+	     "sgtk_char_to_rep" "sgtk_valid_char")
+
+(define-type 'bool "int" "sgtk_rep_to_bool" "sgtk_bool_to_rep" nil)
+
+(define-type 'int "gint" "sgtk_rep_to_int" "sgtk_int_to_rep"
+	     "sgtk_valid_int" '(listable . t))
+
+(define-type 'uint "guint" "sgtk_rep_to_uint" "sgtk_uint_to_rep"
+	     "sgtk_valid_uint" '(listable . t))
+
+(define-type 'long "glong" "sgtk_rep_to_long"
+	     "sgtk_long_to_rep" "sgtk_valid_long")
+
+(define-type 'ulong "gulong" "sgtk_rep_to_ulong"
+	     "sgtk_ulong_to_rep" "sgtk_valid_ulong")
+
+(define-type 'float "gfloat" "sgtk_rep_to_float"
+	     "sgtk_float_to_rep" "sgtk_valid_float")
+
+(define-type 'string "char*" "sgtk_rep_to_string"
+	     "sgtk_string_to_rep" "sgtk_valid_string" '(listable . t))
+
+(define-type 'enum output-complex-type output-rep-to-enum
+	     output-enum-to-rep output-enum-pred)
+
+(define-type 'flags output-complex-type output-rep-to-flags
 	      output-flags-to-rep output-flags-pred)
-	(list 'boxed output-complex-type output-rep-to-boxed
-	      output-boxed-to-rep output-boxed-pred '(listable . t))
-	(list 'pointer "gpointer" "sgtk_rep_to_pointer" "sgtk_pointer_to_rep"
-	      "sgtk_pointerp")
-	(list 'object output-complex-type output-rep-to-object
-	      output-object-to-rep output-object-pred '(listable . t))
-	(list 'static_string "char*" nil "sgtk_static_string_to_rep" nil
-	      '(listable . t))
-	(list 'full-callback "repv" output-rep-to-full-callback nil
-	      "sgtk_valid_function" (cons 'c2args output-full-callback-args))
-	(list 'file-descriptor "int" "sgtk_rep_to_fd" nil "sgtk_valid_fd")
-	(list 'list "GList*" output-rep-to-list output-list-to-rep
-	      output-cvec-pred (cons 'finish output-list-finish))
-	(list 'slist "GSList*" output-rep-to-list output-list-to-rep
-	      output-cvec-pred (cons 'finish output-list-finish))
-	(list 'cvec "sgtk_cvec" output-rep-to-cvec output-cvec-to-rep
-	      output-cvec-pred (cons 'finish output-cvec-finish)
-	      (cons 'c2args output-cvec-args))
-	(list 'cvecr "sgtk_cvec" output-rep-to-cvec output-cvec-to-rep
-	      output-cvec-pred (cons 'finish output-cvec-finish)
-	      (cons 'c2args output-cvec-args))
-	(list 'fvec "sgtk_cvec" output-rep-to-cvec output-cvec-to-rep
-	      output-cvec-pred (cons 'finish output-cvec-finish)
-	      (cons 'c2args output-cvec-args))
-	(list 'ret "sgtk_cvec" output-rep-to-cvec output-cvec-to-rep
-	      output-cvec-pred (cons 'finish output-cvec-finish)
-	      (cons 'c2args output-cvec-args))
-	;;(list 'double "gdouble" "sgtk_rep_to_double" "sgtk_double_to_rep"
-	;;      "sgtk_doublep")
-	(list 'point "GdkPoint" "sgtk_rep_to_point" "sgtk_point_to_rep"
-	      "sgtk_valid_point")
-	(list 'SCM "repv" "" "" nil)))
+
+(define-type 'boxed output-complex-type output-rep-to-boxed
+	     output-boxed-to-rep output-boxed-pred '(listable . t))
+
+(define-type 'pointer "gpointer" "sgtk_rep_to_pointer"
+	     "sgtk_pointer_to_rep" "sgtk_valid_pointer")
+
+(define-type 'object output-complex-type output-rep-to-object
+	     output-object-to-rep output-object-pred '(listable . t))
+
+(define-type 'static_string "const char*" nil
+	     "sgtk_static_string_to_rep" nil '(listable . t))
+
+(define-type 'full-callback "sgtk_protshell*" output-rep-to-full-callback nil
+	     "sgtk_valid_function" (cons 'c2args output-full-callback-args)
+	     (cons 'finish output-full-callback-finish))
+
+(define-type 'file-descriptor "int" "sgtk_rep_to_fd"
+	     "sgtk_fd_to_rep" "sgtk_valid_fd")
+
+(define-type 'list "GList*" output-rep-to-list output-list-to-rep
+	     output-cvec-pred (cons 'finish output-list-finish))
+
+(define-type 'slist "GSList*" output-rep-to-list output-list-to-rep
+	     output-cvec-pred (cons 'finish output-list-finish))
+
+(define-type 'cvec "sgtk_cvec" output-rep-to-cvec output-cvec-to-rep
+	     output-cvec-pred (cons 'finish output-cvec-finish)
+	     (cons 'c2args output-cvec-args))
+
+(define-type 'cvecr "sgtk_cvec" output-rep-to-cvec output-cvec-to-rep
+	     output-cvec-pred (cons 'finish output-cvec-finish)
+	     (cons 'c2args output-cvec-args))
+
+(define-type 'fvec "sgtk_cvec" output-rep-to-cvec output-cvec-to-rep
+	     output-cvec-pred (cons 'finish output-cvec-finish)
+	     (cons 'c2args output-cvec-args))
+
+(define-type 'tvec "sgtk_cvec" output-rep-to-cvec output-cvec-to-rep
+	     output-cvec-pred (cons 'finish output-cvec-finish)
+	     (cons 'c2args output-cvec-args))
+
+(define-type 'ret "sgtk_cvec" output-rep-to-cvec output-cvec-to-rep
+	     output-cvec-pred (cons 'finish output-cvec-finish)
+	     (cons 'c2args output-cvec-args))
+
+(define-type 'double "gdouble" "sgtk_rep_to_double"
+	     "sgtk_double_to_rep" "sgtk_valid_double")
+
+(define-type 'point "GdkPoint" "sgtk_rep_to_point"
+	     "sgtk_point_to_rep" "sgtk_valid_point")
+
+(define-type 'rect "GdkRectangle" "sgtk_rep_to_rect"
+	     "sgtk_rect_to_rep" "sgtk_valid_rect")
+
+(define-type 'SCM "repv" "" "" nil)
