@@ -1,7 +1,7 @@
 /* Copyright (C) 1997, 1998, 1999 Marius Vollmer
  * Copyright (C) 1999 John Harper <john@dcs.warwick.ac.uk>
  *
- * $Id: rep-gtk.c,v 1.14 1999/12/02 18:41:36 john Exp $
+ * $Id: rep-gtk.c,v 1.32 2000/04/09 17:32:28 john Exp $
  *
  * This program is free software; you can redistribute it and/or modify
  * it under the terms of the GNU General Public License as published by
@@ -24,11 +24,19 @@
 #include <gdk/gdkprivate.h>
 #include "rep-gtk.h"
 #include <string.h>
+#include <limits.h>
 
 /* Define this to enable some output during GC and other interesting
    actions. */
 #undef DEBUG_PRINT
 
+static int
+list_length (repv list)
+{
+    repv len = Flength (list);
+    return (len && rep_INTP (len)) ? rep_INT (len) : 0;
+}
+
 
 
 /* Associating SCM values with Gtk pointers.
@@ -79,176 +87,6 @@ forget_proxy (gpointer obj)
 
 
 
-/* Hacking the basic types --jsh */
-
-static inline int
-valid_int_type (repv obj)
-{
-    return rep_INTP (obj) || rep_LONG_INTP (obj);
-}
-
-int
-sgtk_valid_int (repv obj)
-{
-    return valid_int_type (obj);
-}
-
-int
-sgtk_valid_uint (repv obj)
-{
-    return valid_int_type (obj);
-}
-
-int
-sgtk_valid_long (repv obj)
-{
-    return valid_int_type (obj);
-}
-
-int
-sgtk_valid_ulong (repv obj)
-{
-    return valid_int_type (obj);
-}
-
-int
-sgtk_valid_char (repv obj)
-{
-    return rep_INTP (obj);
-}
-
-repv
-sgtk_uint_to_rep (u_long x)
-{
-    if (x <= rep_LISP_MAX_INT)
-	return rep_MAKE_INT (x);
-    else
-	return rep_MAKE_LONG_INT (x);
-}
-
-repv
-sgtk_int_to_rep (long x)
-{
-    if (x <= rep_LISP_MAX_INT && x >= rep_LISP_MIN_INT)
-	return rep_MAKE_INT (x);
-    else
-	return rep_MAKE_LONG_INT (x);
-}
-
-guint
-sgtk_rep_to_uint (repv obj)
-{
-    if (rep_INTP (obj))
-	return rep_INT (obj);
-    else
-	return rep_LONG_INT (obj);
-}
-
-gint
-sgtk_rep_to_int (repv obj)
-{
-    if (rep_INTP (obj))
-	return rep_INT (obj);
-    else
-	return rep_LONG_INT (obj);
-}
-
-gulong
-sgtk_rep_to_ulong (repv obj)
-{
-    if (rep_INTP (obj))
-	return rep_INT (obj);
-    else
-	return rep_LONG_INT (obj);
-}
-
-glong
-sgtk_rep_to_long (repv obj)
-{
-    if (rep_INTP (obj))
-	return rep_INT (obj);
-    else
-	return rep_LONG_INT (obj);
-}
-
-gchar
-sgtk_rep_to_char (repv obj)
-{
-    return rep_INT (obj);
-}
-
-char *
-sgtk_rep_to_string (repv obj)
-{
-    return rep_STRINGP (obj) ? rep_STR (obj) : (u_char *)"";
-}
-
-repv
-sgtk_string_to_rep (char *x)
-{
-    repv obj;
-    if (x == 0)
-	return Qnil;
-    obj = rep_string_dup (x);
-    g_free (x);
-    return obj;
-}
-
-repv
-sgtk_static_string_to_rep (char *x)
-{
-    repv obj;
-    if (x == 0)
-	return Qnil;
-    obj = rep_string_dup (x);
-    return obj;
-}
-
-int
-sgtk_valid_string (repv obj)
-{
-    return rep_STRINGP (obj);
-}
-
-repv
-sgtk_bool_to_rep (int x)
-{
-    return x ? Qt : Qnil;
-}
-
-int
-sgtk_rep_to_bool (repv obj)
-{
-    return obj != Qnil;
-}
-
-int
-sgtk_valid_function (repv obj)
-{
-    return Ffunctionp (obj) != Qnil;
-}
-
-int
-sgtk_valid_fd (repv obj)
-{
-    return rep_FILEP (obj) && rep_LOCAL_FILE_P (obj);
-}
-
-int
-sgtk_rep_to_fd (repv obj)
-{
-    return fileno (rep_FILE(obj)->file.fh);
-}
-
-static int
-list_length (repv list)
-{
-    repv len = Flength (list);
-    return (len && rep_INTP (len)) ? rep_INT (len) : 0;
-}
-
-
-
 /* Storing additional info about a GtkType.
 
    Each GtkType has a unique sequence number.  We use that to simply
@@ -303,7 +141,7 @@ typedef struct _type_infos {
 
 static type_infos *all_type_infos;
 
-/* Finds types that are mentioned in our *.defs files but are not
+/* Find types that are mentioned in our *.defs files but are not
    provided by the Gtk run-time system.  This is only used
    occasionally to update the table in sgtk_try_missing_type.  */
 #ifdef NEED_UNUSED_CODE
@@ -338,6 +176,15 @@ sgtk_register_type_infos (sgtk_type_info **infos)
 #endif
 }
 
+void
+sgtk_register_type_infos_gtk (GtkTypeInfo **infos)
+{
+  GtkTypeInfo **t;
+
+  for (t = infos; t && *t; t++)
+    gtk_type_unique (GTK_TYPE_BOXED, *t);
+}
+
 /* When INFO refers to one of the known `missing' types, we initialize
    that type ourselves.  This is used to fix certain discrepancies
    between old Gtk versions and our *.defs files.  It is not OK to do
@@ -496,6 +343,37 @@ sgtk_find_type_info (GtkType type)
 
   */
 
+struct sgtk_protshell {
+  repv object;
+  struct sgtk_protshell *next;
+  struct sgtk_protshell **prevp;
+};
+
+static GMemChunk *sgtk_protshell_chunk;
+
+/* Analogous to the PROTECTS list of a proxy but for SCM values that
+   are not associated with a particular GtkObject. */
+
+static struct sgtk_protshell *global_protects;
+
+void
+sgtk_unprotect (sgtk_protshell *prot)
+{
+  if ((*prot->prevp = prot->next) != 0)
+    prot->next->prevp = prot->prevp;
+  g_chunk_free (prot, sgtk_protshell_chunk);
+}
+
+static void
+sgtk_mark_protects (sgtk_protshell *prots)
+{
+  while (prots)
+    {
+      rep_MARKVAL (prots->object);
+      prots = prots->next;
+    }
+}
+
 /* The CDR of a GtkObject smob points to one of these.  PROTECTS is a
    Scheme list of all SCM values that need to be protected from the GC
    because they are in use by OBJ.  PROTECTS includes the smob cell
@@ -506,7 +384,7 @@ sgtk_find_type_info (GtkType type)
 typedef struct _sgtk_object_proxy {
   repv car;
   GtkObject *obj;
-  repv protects;
+  struct sgtk_protshell *protects;
   int traced_refs;
   struct _sgtk_object_proxy *next;
 } sgtk_object_proxy;
@@ -515,10 +393,28 @@ typedef struct _sgtk_object_proxy {
 
 static sgtk_object_proxy *all_proxies = NULL;
 
-/* Analogous to the PROTECTS list of a proxy but for SCM values that
-   are not associated with a particular GtkObject. */
+/* Insert the list of protshells starting at PROTS into the global
+   protects list.  This is used when a proxy is freed so that we don't
+   forget about its protects. */
 
-static repv global_protects;
+static void
+sgtk_move_prots_to_global (sgtk_protshell *prots)
+{
+  if (prots)
+    {
+      sgtk_protshell *g = global_protects;
+      global_protects = prots;
+      global_protects->prevp = &global_protects;
+      if (g)
+	{
+	  sgtk_protshell *p;
+	  for (p = prots; p->next; p = p->next)
+	    ;
+	  p->next = g;
+	  g->prevp = &p->next;
+	}
+    }
+}
 
 /* The smob for GtkObjects.  */
 
@@ -527,6 +423,44 @@ static long tc16_gtkobj;
 #define GTKOBJP(x)       (rep_CELL16_TYPEP(x, tc16_gtkobj))
 #define GTKOBJ_PROXY(x)  ((sgtk_object_proxy *)rep_PTR(x))
 
+void
+sgtk_set_protect (repv protector, sgtk_protshell *prot)
+{
+  sgtk_protshell **prevp;
+
+  if (GTKOBJP (protector))
+    prevp = &(GTKOBJ_PROXY(protector)->protects);
+  else
+    prevp = &global_protects;
+  
+  if ((prot->next = *prevp) != 0)
+	prot->next->prevp = &prot->next;
+  *prevp = prot;
+  prot->prevp = prevp;
+}
+
+repv
+sgtk_get_protect (sgtk_protshell *prot)
+{
+  return prot->object;
+}
+
+sgtk_protshell *
+sgtk_new_protect (repv obj)
+{
+  sgtk_protshell *prot = g_chunk_new (sgtk_protshell, sgtk_protshell_chunk);
+  prot->object = obj;
+  return prot;
+}
+
+sgtk_protshell *
+sgtk_protect (repv protector, repv obj)
+{
+  sgtk_protshell *prot = sgtk_new_protect (obj);
+  sgtk_set_protect (protector, prot);
+  return prot;
+}
+
 static void
 mark_traced_ref (GtkWidget *obj, void *data)
 {
@@ -538,7 +472,7 @@ mark_traced_ref (GtkWidget *obj, void *data)
       fprintf (stderr, "marking trace %p %s\n",
 	       proxy->obj, gtk_type_name (GTK_OBJECT_TYPE (proxy->obj)));
 #endif
-      rep_MARKVAL (proxy->protects);
+      sgtk_mark_protects (proxy->protects);
     }
 }
 
@@ -554,7 +488,7 @@ gtkobj_mark (repv obj)
 
   if (GTK_IS_CONTAINER (proxy->obj))
     gtk_container_foreach (GTK_CONTAINER(proxy->obj), mark_traced_ref, NULL);
-  rep_MARKVAL(proxy->protects);
+  sgtk_mark_protects (proxy->protects);
 }
 
 static void
@@ -584,6 +518,7 @@ gtkobj_free (repv obj)
 
   forget_proxy (proxy->obj);
   gtk_object_unref (proxy->obj);
+  sgtk_move_prots_to_global (proxy->protects);
   rep_FREE_CELL ((char *)proxy);
 }
 
@@ -607,53 +542,6 @@ gtkobj_sweep (void)
   }
 }
 
-/* Protect OBJ from being collected.  When PROTECTOR is a GtkObject
-   proxy, OBJ is only protected as long as GtkObject is live. */
-
-repv
-sgtk_protect (repv protector, repv obj)
-{
-  if (GTKOBJP (protector))
-    {
-      sgtk_object_proxy *proxy = GTKOBJ_PROXY(protector);
-      proxy->protects = Fcons (obj, proxy->protects);
-    }
-  else
-    global_protects = Fcons (obj, global_protects);
-  return obj;
-}
-
-static void
-sgtk_unprotect_1 (repv *prev, repv obj)
-{
-  repv walk;
-
-  for (walk = *prev; rep_CONSP (walk);
-       walk = rep_CDR (walk))
-    {
-      if (rep_CAR (walk) == obj)
-	{
-	  *prev = rep_CDR (walk);
-	  break;
-	}
-      else
-	prev = rep_CDRLOC (walk);
-    }
-}
-
-/* XXX - performance improvement by searching only a single
-   proxy->protects.  */
-
-void
-sgtk_unprotect (repv obj)
-{
-  sgtk_object_proxy *proxy;
-
-  for (proxy = all_proxies; proxy; proxy = proxy->next)
-    sgtk_unprotect_1 (&proxy->protects, obj);
-  sgtk_unprotect_1 (&global_protects, obj);
-}
-
 /* Treating GtkObject proxies right during GC.  We need to run custom
    code during the mark phase of the Scheme GC.  We do this by
    creating a new smob type and allocating one actual smob of it.
@@ -718,11 +606,15 @@ gtkobj_marker_hook (void)
 	  fprintf (stderr, "hooking %p %s\n",
 		   proxy->obj, gtk_type_name (GTK_OBJECT_TYPE (proxy->obj)));
 #endif
-	  rep_MARKVAL (proxy->protects);
+	  /* mark the proxy itself */
+	  rep_MARKVAL (rep_VAL (proxy));
 	}
+      /* always mark the protected objects, since they're moved to
+         the global_protects list if the object is freed */
+      sgtk_mark_protects (proxy->protects);
       proxy->traced_refs = 0;
     }
-  rep_MARKVAL (global_protects);
+  sgtk_mark_protects (global_protects);
 }
 
 /* Create a proxy for OBJ. */
@@ -740,7 +632,7 @@ make_gtkobj (GtkObject *obj)
 	   gtk_type_name (GTK_OBJECT_TYPE (obj)));
 #endif
   proxy->obj = obj;
-  proxy->protects = Qnil;
+  proxy->protects = NULL;
   proxy->traced_refs = 0;
   proxy->next = all_proxies;
   all_proxies = proxy;
@@ -748,7 +640,6 @@ make_gtkobj (GtkObject *obj)
   proxy->car = tc16_gtkobj;
   enter_proxy (obj, rep_VAL(proxy));
 
-  sgtk_protect (rep_VAL(proxy), rep_VAL(proxy));
   return rep_VAL(proxy);
 }
 
@@ -893,7 +784,7 @@ sgtk_rep_to_flags (repv obj, sgtk_enum_info *info)
 {
   int ans = 0;
 
-  while (obj != Qnil)
+  while (rep_CONSP(obj) && !rep_INTERRUPTP)
     {
       int i;
       repv sym = rep_CAR (obj);
@@ -906,6 +797,7 @@ sgtk_rep_to_flags (repv obj, sgtk_enum_info *info)
 	    break;
 	  }
       obj = rep_CDR (obj);
+      rep_TEST_INT;
     }
   
   return ans;
@@ -1074,48 +966,6 @@ sgtk_valid_boxed (repv obj, sgtk_boxed_info *info)
   return (BOXED_P (obj) && BOXED_INFO (obj) == info);
 }
 
-/* Floats.
-
-   Only here to set things straight.
-
-   Of course, rep currently only has integer. Kludge this. --jsh */
-
-int
-sgtk_valid_float (repv obj)
-{
-  return rep_INTP (obj);
-}
-
-gfloat
-sgtk_rep_to_float (repv obj)
-{
-  return rep_INTP (obj) ? rep_INT (obj) : 0.0;
-}
-
-repv
-sgtk_float_to_rep (gfloat f)
-{
-  return rep_MAKE_INT ((long int)f);
-}
-
-int
-sgtk_valid_double (repv obj)
-{
-  return rep_INTP (obj);
-}
-
-double
-sgtk_rep_to_double (repv obj)
-{
-  return rep_INTP (obj) ? rep_INT (obj) : 0.0;
-}
-
-repv
-sgtk_double_to_rep (double f)
-{
-  return rep_MAKE_INT ((long int)f);
-}
-
 int
 sgtk_valid_point (repv obj)
 {
@@ -1140,6 +990,34 @@ sgtk_point_to_rep (GdkPoint p)
 		sgtk_int_to_rep (p.y));
 }
 
+int
+sgtk_valid_rect (repv obj)
+{
+  return rep_CONSP (obj)
+    && sgtk_valid_point (rep_CAR (obj))
+    && sgtk_valid_point (rep_CDR (obj));
+}
+
+GdkRectangle
+sgtk_rep_to_rect (repv obj)
+{
+  GdkRectangle res;
+  res.x = rep_INT (rep_CAAR (obj));
+  res.y = rep_INT (rep_CDAR (obj));
+  res.width = rep_INT (rep_CADR (obj));
+  res.height = rep_INT (rep_CDDR (obj));
+  return res;
+}
+
+repv
+sgtk_rect_to_rep (GdkRectangle r)
+{
+  return Fcons (Fcons (rep_MAKE_INT (r.x),
+		       rep_MAKE_INT (r.y)),
+		Fcons (rep_MAKE_INT (r.width),
+		       rep_MAKE_INT (r.height)));
+}
+
 
 
 /* GtkType objects
@@ -1200,288 +1078,6 @@ sgtk_type_to_rep (GtkType t)
   return sgtk_uint_to_rep (t);
 }
 
-
-
-/* Composites. */
-
-int
-sgtk_valid_composite (repv obj, int (*predicate)(repv))
-{
-  return sgtk_valid_complen (obj, predicate, -1);
-}
-
-int
-sgtk_valid_complen (repv obj, int (*predicate)(repv), int len)
-{
-  int actual_len;
-
-  if (rep_LISTP(obj))
-    {
-      actual_len = list_length (obj);
-
-      if (len >= 0 && len != actual_len)
-	return 0;
-
-      if (predicate)
-	{
-	  while (rep_CONSP(obj))
-	    {
-	      if (!predicate (rep_CAR(obj)))
-		return 0;
-	      obj = rep_CDR(obj);
-	    }
-	}
-      return 1;
-    }
-  else if (rep_VECTORP(obj))
-    {
-      int i;
-      repv *elts;
-
-      actual_len = rep_VECT_LEN (obj);
-      if (len >= 0 && len != actual_len)
-	return 0;
-
-      if (predicate)
-	{
-	  elts = rep_VECT(obj)->array;
-	  for (i = 0; i < len; i++)
-	    if (!predicate(elts[i]))
-	      return 0;
-	}
-      return 1;
-    }
-  else
-    return 0;
-}
-
-repv
-sgtk_slist_to_rep (GSList *list, repv (*toscm)(void*))
-{
-  repv res, *tail = &res;
-  while (list)
-    {
-      *tail = Fcons (toscm (&list->data), *tail);
-      tail = rep_CDRLOC (*tail);
-      list = list->next;
-    }
-  *tail = Qnil;
-  return res;
-}
-
-GSList*
-sgtk_rep_to_slist (repv obj, void (*fromscm)(repv, void*))
-{
-  GSList *res, **tail = &res;
-
-  if (obj == Qnil || (rep_CONSP(obj)))
-    {
-      while (rep_CONSP(obj))
-	{
-	  *tail = g_slist_alloc ();
-	  if (fromscm)
-	    fromscm (rep_CAR (obj), &(*tail)->data);
-	  obj = rep_CDR(obj);
-	  tail = &(*tail)->next;
-	}
-    }
-  else if (rep_VECTORP(obj))
-    {
-      int len = rep_VECT_LEN (obj), i;
-      repv *elts = rep_VECT(obj)->array;
-      for (i = 0; i < len; i++)
-	{
-	  *tail = g_slist_alloc ();
-	  if (fromscm)
-	    fromscm (elts[i], &(*tail)->data);
-	  tail = &(*tail)->next;
-	}
-    }
-  *tail = NULL;
-  return res;
-}
-
-void
-sgtk_slist_finish (GSList *list, repv obj, repv (*toscm)(void*))
-{
-  if (toscm)
-    {
-      if (obj == Qnil || (rep_CONSP(obj)))
-	{
-	  while (rep_CONSP(obj) && list)
-	    {
-	      rep_CAR(obj) = toscm (list->data);
-	      obj = rep_CDR(obj);
-	      list = list->next;
-	    }
-	}
-      else if (rep_VECTORP(obj))
-	{
-	  int len = rep_VECT_LEN (obj), i;
-	  repv *elts = rep_VECT(obj)->array;
-	  for (i = 0; i < len && list; i++)
-	    {
-	      elts[i] = toscm (list->data);
-	      list = list->next;
-	    }
-	}
-    }
-
-  g_slist_free (list);
-}
-
-repv
-sgtk_list_to_rep (GList *list, repv (*toscm)(void*))
-{
-  repv res, *tail = &res;
-  while (list)
-    {
-      *tail = Fcons (toscm (&list->data), *tail);
-      tail = rep_CDRLOC (*tail);
-      list = list->next;
-    }
-  *tail = Qnil;
-  return res;
-}
-
-GList*
-sgtk_rep_to_list (repv obj, void (*fromscm)(repv, void*))
-{
-  GList *res = NULL, *tail;
-  
-  if (obj == Qnil || (rep_CONSP(obj)))
-    {
-      while (rep_CONSP(obj))
-      {
-        GList *n = g_list_alloc ();
-	if (res == NULL)
-	  res = tail = n;
-	else 
-	  {
-	    g_list_concat (tail, n);
-	    tail = n;
-	  }
-	if (fromscm)
-	  fromscm (rep_CAR (obj), &(n->data));
-	obj = rep_CDR(obj);
-      }
-    }
-  else if (rep_VECTORP(obj))
-    {
-      int len = rep_VECT_LEN (obj), i;
-      repv *elts = rep_VECT(obj)->array;
-      for (i = 0; i < len; i++)
-	{
-	  GList *n = g_list_alloc ();
-	  if (res == NULL)
-	    res = tail = n;
-	  else 
-	    {
-	      g_list_concat (tail, n);
-	      tail = n;
-	    }
-	  if (fromscm)
-	    fromscm (elts[i], &(n->data));
-	}
-    }
-  return res;
-}
-
-void
-sgtk_list_finish (GList *list, repv obj, repv (*toscm)(void*))
-{
-  if (toscm)
-    {
-      if (obj == Qnil || (rep_CONSP(obj)))
-	{
-	  while (rep_CONSP(obj) && list)
-	    {
-	      rep_CAR (obj) = toscm (list->data);
-	      obj = rep_CDR(obj);
-	      list = list->next;
-	    }
-	}
-      else if (rep_VECTORP(obj))
-	{
-	  int len = rep_VECT_LEN (obj), i;
-	  repv *elts = rep_VECT(obj)->array;
-	  for (i = 0; i < len && list; i++)
-	    {
-	      elts[i] = toscm (list->data);
-	      list = list->next;
-	    }
-	}
-    }
-  
-  g_list_free (list);
-}
-
-sgtk_cvec
-sgtk_rep_to_cvec (repv obj, void (*fromscm)(repv, void*), size_t sz)
-{
-  sgtk_cvec res;
-  int i;
-  char *ptr;
-
-  if (rep_LISTP(obj))
-    {
-      res.count = list_length (obj);
-      res.vec = rep_alloc (res.count * sz);
-      if (fromscm)
-	{
-	  for (i = 0, ptr = res.vec; i < res.count; i++, ptr += sz)
-	    {
-	      fromscm (rep_CAR (obj), ptr);
-	      obj = rep_CDR(obj);
-	    }
-	}
-    }
-  else if (rep_VECTORP(obj))
-    {
-      repv *elts = rep_VECT(obj)->array;
-      res.count = rep_VECT_LEN (obj);
-      res.vec = rep_alloc (res.count * sz);
-      if (fromscm)
-	{
-	  for (i = 0, ptr = res.vec; i < res.count; i++, ptr += sz)
-	    fromscm (elts[i], ptr);
-	}
-    }
-
-  return res;
-}
-
-void
-sgtk_cvec_finish (sgtk_cvec *cvec, repv obj, repv (*toscm)(void *), size_t sz)
-{
-  if (toscm)
-    {
-      if (obj == Qnil || (rep_CONSP(obj)))
-	{
-	  int i, len = cvec->count;
-	  char *ptr;
-
-	  for (i = 0, ptr = cvec->vec;
-	       i < len && rep_CONSP(obj);
-	       i++, ptr += sz, obj = rep_CDR (obj))
-	    {
-	      rep_CAR (obj) = toscm (ptr);
-	    }
-	}
-      else if (rep_VECTORP(obj))
-	{
-	  repv *elts = rep_VECT(obj)->array;
-	  int len1 = rep_VECT_LEN (obj), len2 = cvec->count, i;
-	  char *ptr;
-
-	  for (i = 0, ptr = cvec->vec; i < len1 && i < len2; i++, ptr += sz)
-	    elts[i] = toscm (ptr);
-	}
-    }
-
-  rep_free (cvec->vec);
-}
-
 
 /* converting between SCM and GtkArg */
 
@@ -1542,16 +1138,17 @@ sgtk_valid_arg (GtkArg *a, repv obj)
     case GTK_TYPE_NONE:
       return TRUE;
     case GTK_TYPE_CHAR:
-      return rep_INTP(obj);
+      return sgtk_valid_char(obj);
     case GTK_TYPE_BOOL:
       return TRUE;
     case GTK_TYPE_INT:
     case GTK_TYPE_UINT:
     case GTK_TYPE_LONG:
     case GTK_TYPE_ULONG:
+      return sgtk_valid_int (obj);
     case GTK_TYPE_FLOAT:
     case GTK_TYPE_DOUBLE:
-      return rep_INTP (obj);		/* XXX to be fixed.. --jsh */
+      return sgtk_valid_float (obj);
     case GTK_TYPE_STRING:
       return rep_STRINGP (obj);
     case GTK_TYPE_ENUM:
@@ -1565,7 +1162,7 @@ sgtk_valid_arg (GtkArg *a, repv obj)
 				     sgtk_find_type_info (a->type)));
       break;
     case GTK_TYPE_CALLBACK:
-      return Ffunctionp (obj);
+      return sgtk_valid_function (obj);
     case GTK_TYPE_OBJECT:
       return sgtk_is_a_gtkobj (a->type, obj);
     default:
@@ -1588,16 +1185,16 @@ sgtk_rep_to_arg (GtkArg *a, repv obj, repv protector)
       GTK_VALUE_BOOL(*a) = obj != Qnil;
       break;
     case GTK_TYPE_INT:
-      GTK_VALUE_INT(*a) = rep_INT (obj);
+      GTK_VALUE_INT(*a) = sgtk_rep_to_int (obj);
       break;
     case GTK_TYPE_UINT:
-      GTK_VALUE_UINT(*a) = rep_INT (obj);
+      GTK_VALUE_UINT(*a) = sgtk_rep_to_uint (obj);
       break;
     case GTK_TYPE_LONG:
-      GTK_VALUE_LONG(*a) = rep_INT (obj);
+      GTK_VALUE_LONG(*a) = sgtk_rep_to_long (obj);
       break;
     case GTK_TYPE_ULONG:
-      GTK_VALUE_ULONG(*a) = rep_INT (obj);
+      GTK_VALUE_ULONG(*a) = sgtk_rep_to_ulong (obj);
       break;
     case GTK_TYPE_FLOAT:
       GTK_VALUE_FLOAT(*a) = sgtk_rep_to_float (obj);
@@ -1606,7 +1203,7 @@ sgtk_rep_to_arg (GtkArg *a, repv obj, repv protector)
       GTK_VALUE_DOUBLE(*a) = sgtk_rep_to_double (obj);
       break;
     case GTK_TYPE_STRING:
-      GTK_VALUE_STRING(*a) = rep_STR (obj);
+      GTK_VALUE_STRING(*a) = sgtk_rep_to_string (obj);
       break;
     case GTK_TYPE_ENUM:
       GTK_VALUE_ENUM(*a) =
@@ -1648,16 +1245,16 @@ sgtk_rep_to_ret (GtkArg *a, repv obj)
       *GTK_RETLOC_BOOL(*a) = (obj != Qnil);
       break;
     case GTK_TYPE_INT:
-      *GTK_RETLOC_INT(*a) = rep_INT (obj);
+      *GTK_RETLOC_INT(*a) = sgtk_rep_to_int (obj);
       break;
     case GTK_TYPE_UINT:
-      *GTK_RETLOC_UINT(*a) = rep_INT (obj);
+      *GTK_RETLOC_UINT(*a) = sgtk_rep_to_uint (obj);
       break;
     case GTK_TYPE_LONG:
-      *GTK_RETLOC_LONG(*a) = rep_INT (obj);
+      *GTK_RETLOC_LONG(*a) = sgtk_rep_to_long (obj);
       break;
     case GTK_TYPE_ULONG:
-      *GTK_RETLOC_ULONG(*a) = rep_INT (obj);
+      *GTK_RETLOC_ULONG(*a) = sgtk_rep_to_ulong (obj);
       break;
     case GTK_TYPE_FLOAT:
       *GTK_RETLOC_FLOAT(*a) = sgtk_rep_to_float (obj);
@@ -1731,14 +1328,43 @@ DEFUN ("gtk-callback-trampoline", Fgtk_callback_trampoline,
    scm_gc_heap_lock is set during gc.  */
 #define rep_GC_P (rep_in_gc)
 
+struct callback_info {
+  GtkObject *obj;
+  repv proc;
+  gint n_args;
+  GtkArg *args;
+};
+
+static repv
+inner_callback_marshal (repv data)
+{
+  struct callback_info *info = (struct callback_info *) rep_PTR (data);
+  int i;
+  repv args = Qnil, ans;
+
+  for (i = info->n_args-1; i >= 0; i--)
+    args = Fcons (sgtk_arg_to_rep (info->args+i, 0), args);
+  args = Fcons (sgtk_wrap_gtkobj (info->obj), args);
+
+  if (rep_CAR(callback_trampoline) == Qnil)
+    ans = rep_funcall (info->proc, args, rep_FALSE);
+  else
+    ans = rep_funcall (rep_CAR(callback_trampoline),
+		       Fcons (info->proc, Fcons (args, Qnil)), rep_FALSE);
+
+  if (info->args[info->n_args].type != GTK_TYPE_NONE)
+    sgtk_rep_to_ret (info->args+info->n_args, ans);
+
+  return Qnil;
+}
+
 void
 sgtk_callback_marshal (GtkObject *obj,
 		       gpointer data,
 		       guint n_args,
 		       GtkArg *args)
 {
-  repv real_args = Qnil, ans;
-  int i;
+  struct callback_info info;
 
   if (rep_GC_P)
     {
@@ -1748,27 +1374,21 @@ sgtk_callback_marshal (GtkObject *obj,
       return;
     }
   
-  for (i = n_args - 1; i >= 0; i--)
-    real_args = Fcons (sgtk_arg_to_rep (args + i, 0), real_args);
+  info.obj = obj;
+  info.proc = ((sgtk_protshell *)data)->object;
+  info.n_args = n_args;
+  info.args = args;
 
-  real_args = Fcons (sgtk_wrap_gtkobj (obj), real_args);
-
-  if (rep_CAR(callback_trampoline) == Qnil)
-    ans = rep_funcall ((repv)data, real_args, rep_FALSE);
-  else
-    ans = rep_funcall (rep_CAR(callback_trampoline),
-		       Fcons ((repv)data, Fcons (real_args, Qnil)), rep_FALSE);
+  rep_call_with_barrier (inner_callback_marshal,
+			 rep_VAL(&info), rep_TRUE, 0, 0, 0);
 
   sgtk_callback_postfix ();
-
-  if (ans && args[n_args].type != GTK_TYPE_NONE)
-    sgtk_rep_to_ret (args + n_args, ans);
 }
 
 void
 sgtk_callback_destroy (gpointer data)
 {
-  sgtk_unprotect ((repv)data);
+  sgtk_unprotect ((sgtk_protshell *)data);
 }
 
 
@@ -2040,8 +1660,8 @@ sgtk_build_args (sgtk_object_info *info, int *n_argsp, repv scm_args,
   return args;
 }
 
-repv
-sgtk_gtk_object_new (repv type_obj, repv scm_args)
+DEFUN("gtk-object-new", Fgtk_object_new, Sgtk_object_new,
+      (repv type_obj, repv scm_args), rep_Subr2)
 {
   int n_args;
   sgtk_object_info *info;
@@ -2067,8 +1687,8 @@ sgtk_gtk_object_new (repv type_obj, repv scm_args)
   return scm_obj;
 }
 
-repv
-sgtk_gtk_object_set (repv scm_obj, repv scm_args)
+DEFUN("gtk-object-set", Fgtk_object_set, Sgtk_object_set,
+      (repv scm_obj, repv scm_args), rep_Subr2)
 {
   int n_args;
   sgtk_object_info *info;
@@ -2092,8 +1712,8 @@ sgtk_gtk_object_set (repv scm_obj, repv scm_args)
   return Qnil;
 }
 
-repv
-sgtk_gtk_object_get (repv scm_obj, repv argsym)
+DEFUN ("gtk-object-get", Fgtk_object_get, Sgtk_object_get,
+       (repv scm_obj, repv argsym), rep_Subr2)
 {
   GtkObject *obj;
   sgtk_object_info *info;
@@ -2229,13 +1849,42 @@ sgtk_signal_emit (GtkObject *obj, char *name, repv scm_args)
    table really.. */
 static GHashTable *input_tags, *input_callbacks;
 
+struct input_callback_data {
+    void (*func)(int);
+    int fd;
+};
+
+struct timeout_data {
+    struct timeout_data *next;
+    int timed_out;
+    int idle_counter;
+    u_long this_timeout_msecs;
+    u_long actual_timeout_msecs;
+    int gtk_tag;
+};
+
+static struct timeout_data *context;
+
+static repv
+inner_input_callback (repv data_)
+{
+    struct input_callback_data *data
+	= (struct input_callback_data *) rep_PTR (data_);
+    (*data->func) (data->fd);
+    return Qnil;
+}
+
 static void
 sgtk_input_callback (gpointer data, gint fd, GdkInputCondition cond)
 {
-    void (*func)(int fd) = g_hash_table_lookup (input_callbacks,
-						(gpointer) fd);
-    if (func != 0)
-	(*func) (fd);
+    struct input_callback_data d;
+    d.func = g_hash_table_lookup (input_callbacks, (gpointer) fd);
+    d.fd = fd;
+    if (d.func != 0)
+    {
+	rep_call_with_barrier (inner_input_callback, rep_VAL(&d),
+			       rep_TRUE, 0, 0, 0);
+    }
     sgtk_callback_postfix ();
 }
 
@@ -2272,54 +1921,105 @@ sgtk_deregister_input_fd (int fd)
     }
 }
 
-/* Mimic the rep idle stuff (every second) using a timeout that is reset
-   after each callback/event. */
-static int idle_timeout_set = 0, idle_timeout_counter = 0, idle_timeout_tag;
-
-static gint
-idle_timeout_callback (gpointer data)
+static gboolean
+timeout_callback (gpointer data)
 {
-    rep_proc_periodically ();
-    rep_on_idle (idle_timeout_counter++);
-    if (rep_INTERRUPTP && gtk_main_level () > 0)
+    struct timeout_data *d = data;
+
+    d->gtk_tag = 0;
+    d->timed_out = 1;
+
+    /* Only quit if we'd return to the correct event loop */
+    if (context == d)
 	gtk_main_quit ();
-    else if (rep_redisplay_fun != 0)
-	(*rep_redisplay_fun)();
-    return 1;
+
+    return FALSE;
 }
 
 static void
-reset_idle_timeout (void)
+unset_timeout (void)
 {
-    if (idle_timeout_set)
-	gtk_timeout_remove (idle_timeout_tag);
-    idle_timeout_counter = 0;
-    idle_timeout_tag = gtk_timeout_add (1000, idle_timeout_callback, 0);
-    idle_timeout_set = TRUE;
+    if (context != 0)
+    {
+	if (context->gtk_tag != 0)
+	    gtk_timeout_remove (context->gtk_tag);
+	context->gtk_tag = 0;
+    }
+}
+
+static void
+set_timeout (void)
+{
+    if (context != 0 && !context->timed_out && !context->gtk_tag)
+    {
+	u_long max_sleep = rep_max_sleep_for ();
+	context->this_timeout_msecs = rep_input_timeout_secs * 1000;
+	context->actual_timeout_msecs = MIN (context->this_timeout_msecs,
+					     max_sleep);
+	context->gtk_tag = gtk_timeout_add (context->actual_timeout_msecs,
+					    timeout_callback,
+					    (gpointer) context);
+    }
 }
 
 /* Call this after executing any callbacks that could invoke Lisp code */
 void
 sgtk_callback_postfix (void)
 {
+    unset_timeout ();
     if (rep_INTERRUPTP && gtk_main_level () > 0)
 	gtk_main_quit ();
     else if (rep_redisplay_fun != 0)
 	(*rep_redisplay_fun)();
-    reset_idle_timeout ();
+    if (context != 0)
+    {
+	context->timed_out = 0;
+	set_timeout ();
+	context->idle_counter = 0;
+    }
 }
 
 /* This function replaces the standard rep event loop. */
 static repv
 sgtk_event_loop (void)
 {
+    struct timeout_data data;
+
+    data.idle_counter = 0;
+    data.gtk_tag = 0;
+    data.next = context;
+    context = &data;
+
     while (1)
     {
+	u_long max_sleep = rep_max_sleep_for ();
+
 	if (rep_redisplay_fun != 0)
 	    (*rep_redisplay_fun)();
 
-	reset_idle_timeout ();
-	gtk_main ();
+	if (max_sleep == 0)
+	{
+	    while (gtk_events_pending ())
+		gtk_main_iteration_do (FALSE);
+	    Fthread_yield ();
+	}
+	else
+	{
+	    data.timed_out = 0;
+	    set_timeout ();
+	    gtk_main ();
+	    unset_timeout ();
+	    if (data.timed_out)
+	    {
+		if (data.actual_timeout_msecs < data.this_timeout_msecs)
+		{
+		    Fthread_suspend (Qnil, rep_MAKE_INT (data.this_timeout_msecs
+							 - data.actual_timeout_msecs));
+		}
+		else
+		    rep_on_idle (data.idle_counter++);
+	    }
+	}
 
 	rep_proc_periodically ();
 
@@ -2328,7 +2028,12 @@ sgtk_event_loop (void)
 	{
 	    repv result;
 	    if(rep_handle_input_exception (&result))
+	    {
+		context = data.next;
+		/* reset the timeout for any containing event loop */
+		set_timeout ();
 		return result;
+	    }
 	}
 
 #ifdef C_ALLOCA
@@ -2363,7 +2068,7 @@ sgtk_set_standalone (int flag)
 }
 
 int
-sgtk_is_standalone ()
+sgtk_is_standalone (void)
 {
   return standalone_p;
 }
@@ -2393,7 +2098,9 @@ sgtk_init_substrate (void)
 				      boxed_sweep, 0, 0,
 				      0, 0, 0, 0, 0, 0);
 
-  global_protects = Qnil;
+  global_protects = NULL;
+  sgtk_protshell_chunk = g_mem_chunk_create (sgtk_protshell, 128,
+					     G_ALLOC_AND_FREE);
   
   callback_trampoline = Fcons (Qnil, Qnil);
   rep_mark_static (&callback_trampoline);
@@ -2410,13 +2117,16 @@ sgtk_init_substrate (void)
 
   rep_ADD_SUBR (Sgtk_callback_trampoline);
   rep_ADD_SUBR (Sgtk_standalone_p);
+  rep_ADD_SUBR (Sgtk_object_new);
+  rep_ADD_SUBR (Sgtk_object_set);
+  rep_ADD_SUBR (Sgtk_object_get);
   rep_INTERN (gtk_major_version);
   rep_INTERN (gtk_minor_version);
   rep_INTERN (gtk_micro_version);
   rep_INTERN (gtk);
-  rep_SYM (Qgtk_major_version)->value = rep_MAKE_INT (GTK_MAJOR_VERSION);
-  rep_SYM (Qgtk_minor_version)->value = rep_MAKE_INT (GTK_MINOR_VERSION);
-  rep_SYM (Qgtk_micro_version)->value = rep_MAKE_INT (GTK_MICRO_VERSION);
+  Fset (Qgtk_major_version, rep_MAKE_INT (GTK_MAJOR_VERSION));
+  Fset (Qgtk_minor_version, rep_MAKE_INT (GTK_MINOR_VERSION));
+  Fset (Qgtk_micro_version, rep_MAKE_INT (GTK_MICRO_VERSION));
 }
 
 static int sgtk_inited = 0;
@@ -2438,7 +2148,10 @@ sgtk_init_with_args (int *argcp, char ***argvp)
     {
       char *tem = getenv ("REP_GTK_DONT_INITIALIZE");
       if (tem == 0 || atoi (tem) == 0)
-	gtk_init (argcp, argvp);
+        {
+	  gtk_set_locale ();
+	  gtk_init (argcp, argvp);
+        }
     }
 
   if (rep_recurse_depth >= 0)
@@ -2492,7 +2205,7 @@ make_argv (repv list, int *argc, char ***argv)
 }
 
 void
-sgtk_init ()
+sgtk_init (void)
 {
   int argc;
   char **argv;
@@ -2501,8 +2214,8 @@ sgtk_init ()
   if (sgtk_inited)
     return;
 
-  make_argv (Fcons (rep_SYM(Qprogram_name)->value,
-		    rep_SYM(Qcommand_line_args)->value), &argc, &argv);
+  make_argv (Fcons (Fsymbol_value (Qprogram_name, Qt),
+		    Fsymbol_value (Qcommand_line_args, Qt)), &argc, &argv);
   sgtk_init_with_args (&argc, &argv);
 
   argc--; argv++;
@@ -2515,22 +2228,29 @@ sgtk_init ()
       argc--;
       argv++;
   }
-  rep_SYM(Qcommand_line_args)->value = head;
+  Fset (Qcommand_line_args, head);
 }
 
 
 
 /* DL hooks */
 
-repv rep_dl_feature;
+extern void sgtk_init_gtk_gtk_glue (void);
 
 repv
 rep_dl_init (void)
 {
+#if rep_INTERFACE >= 9
+  repv tem = rep_push_structure ("gtk");
+#endif
+
   sgtk_init_gtk_gtk_glue ();
 
-  rep_dl_feature = Qgtk;
-  return Qt;
+#if rep_INTERFACE >= 9
+  return rep_pop_structure (tem);
+#else
+  return Qgtk;
+#endif
 }
 
 /* This is required mainly since other dls may try to unregister
